#include  <stdio.h>
/*{{{  #defines*/
#define  TY_KIND_MASK		0x3FL	/* ...can store KIND values < 64 */

#define  TY_INTEGER		0x40L
#define	 TY_REAL		0x80L
#define  TY_DP			0x100L
#define  TY_COMPLEX		0x200L
#define  TY_LOGICAL		0x400L
#define  TY_CHARACTER		0x800L
#define  TY_DERIVED		0x1000L
#define  TY_NONE		0x2000L

#define  TY_BASIC_TYPE	(TY_INTEGER | TY_REAL | TY_DP | TY_COMPLEX |\
			 TY_LOGICAL | TY_CHARACTER | TY_DERIVED | TY_NONE)

#define  TY_TYPE	(TY_KIND_MASK | TY_BASIC_TYPE)


/*----------------------------------------------------------------
'Mapvec' macros
---------------
Each 'mapvec' has the form:

INTEGER  ::  attrs,  a_rank,  e_rank,  array_struct_ptr,  proc_struct_ptr
INTEGER, DIMENSION (a_rank)  ::  l, n, s, segn, al, an, as,  tn,  b, fmt, d
INTEGER, DIMENSION (e_rank)  ::  ep, ed
INTEGER, DIMENSION (3 + 11*a_rank + 2*e_rank)  ::  mapvec

mapvec = (/ attrs,  array_struct_ptr,  proc_struct_ptr,         &
            l, n, s, segn, al, an, as, tn, b, fmt, d, ep, ed /)
----------------------------------------------------------------*/
/*-------------------------*/
/* Flags in mapvec (attrs) */
/*-------------------------*/
#define  MA_ARRAY_RANK_MASK		0x1C000L  /* field for 'array_rank' */
#define  MA_ARRAY_RANK_SHIFT		14	  /* shift for 'array_rank' */

/*------------------------*/
/* Subscripts in 'mapvec' */
/*------------------------*/
#define  l_subscr(a_rank,dim)		(3 + (dim))
#define  n_subscr(a_rank,dim)		(3 + (a_rank) + (dim))
#define  s_subscr(a_rank,dim)		(3 + 2*(a_rank) + (dim))
#define  segn_subscr(a_rank,dim)	(3 + 3*(a_rank) + (dim))
#define  al_subscr(a_rank,dim)		(3 + 4*(a_rank) + (dim))
#define  an_subscr(a_rank,dim)		(3 + 5*(a_rank) + (dim))
#define  as_subscr(a_rank,dim)		(3 + 6*(a_rank) + (dim))
#define  tn_subscr(a_rank,dim)		(3 + 7*(a_rank) + (dim))
#define  b_subscr(a_rank,dim)		(3 + 8*(a_rank) + (dim))
#define  fmt_subscr(a_rank,dim)		(3 + 9*(a_rank) + (dim))
#define  d_subscr(a_rank,dim)		(3 + 10*(a_rank) + (dim))
/*}}}*/

/*{{{  void  ad_alloc_(ptr_p,  type_p,  len_p,  stack_ptr_p,  stack_len_p)--ok*/
void  ad_alloc_(ptr_p,  type_p,  len_p,  stack_ptr_p,  stack_len_p)
/*---------------------------------------------------------------*/
	int  *ptr_p,  *type_p,  *len_p,  *stack_ptr_p,  *stack_len_p;
/*--------------------------------------------------------------*/
/* Forget about (type == CHARACTER or DERIVED), or KIND values,	*/
/* for the time being!						*/
/*--------------------------------------------------------------*/
{
	switch (*type_p) {
	case  TY_DP :
	case  TY_COMPLEX :
		*ptr_p = *stack_ptr_p / 2;
		*stack_ptr_p = *stack_ptr_p + (2 * (*len_p));
		break;
	
	default :
		*ptr_p = *stack_ptr_p;
		*stack_ptr_p = *stack_ptr_p + (2 * ((*len_p) + 1) / 2);
		break;
	}

	if (*stack_ptr_p > *stack_len_p) {
		fprintf (stderr, "** RUNTIME ERROR **  Stack overflow.  Recompile using \'-s\' flag to specify a larger stack size (currently %d)",
		                  *stack_len_p);
		exit (1);
	}
}
/*}}}*/

/*{{{  void  ad_loc_sect_ (eta_p, loc_l_p, loc_n_p, mapvec, dim_p, this_p_p, cycle_p)*/
void  ad_loc_sect_ (eta_p, loc_l_p, loc_n_p, mapvec, dim_p, this_p_p, cycle_p)
/*-------------------------------------------------------------------------*/
	int  *eta_p,  *loc_l_p,  *loc_n_p,  *mapvec,  *dim_p,
	     *this_p_p,  *cycle_p;

/*{{{  < DESCRIPTION >*/
#if 0
/*----------------------------------------------------------------
	This evaluates the (eta, loc_l, loc_n) parameters for a
regular section selected from a *distributed* dimension, 'dim',
of an array whose mapping is given by 'mapvec []'.

The following formulae apply for BLOCK *or* CYCLIC.  In the
latter case, 'cycle' gives the cycle number (== 0 for BLOCK),
and 'this_p' has been adjusted (prior to input) as follows:

	this_p -> this_p + cycle * nprocs

/*{{{  loc_s*/
loc_s
-----
= the section stride in terms of target subscripts.
= (s * as)

N.B.  If this is 0 it's an error.  In that case substitute 1 as
a simple fix!

/*}}}*/
/*{{{  loc_sect_l*/
loc_sect_l
----------
= 'l' in terms of local subscripts w.r.t. target array.	 We must
divide it by |as| to get it in terms of local subscripts.

loc_sect_l = al + (l * as) - (b * this_p)

/*}}}*/
/*{{{  eta*/
eta
---
= # of 1st element of the section in the local segment, or more precisely:
        if (loc_s > 0), 1st element >= lb;
        if (loc_s < 0), 1st element <= ub.

eta = MAX ((loc_s - loc_sect_l - 1) / loc_s, 0)		! loc_s > 0
    = MAX ((loc_s - loc_sect_l + b) / loc_s, 0)		! loc_s < 0

/*}}}*/
/*{{{  loc_l*/
loc_l
-----
= local subscript w.r.t. the target array of element 'eta'.
If it's outside the local segment for reasons (ii) or (iii) above,
then  (loc_l > b-1) if (loc_s > 0) or (loc_l < 0) if (loc_s < 0).
		
loc_l = loc_sect_l + (eta * loc_s)


N.B.
For the return value, i.e. the actual local subscript, we must:
(i)  divide this by |as|:

	loc_l -> loc_l / |as|.

(ii) for CYCLIC distribution, add  (cycle * CEILING (b / |as|)):

	locl -> locl + cycle * CEILING (b / |as|)

/*}}}*/
/*{{{  loc_n*/
loc_n
-----
loc_n = MAX ((loc_u - loc_l + loc_s)/loc_s, 0)

where:

loc_u = MIN (loc_sect_l + (n-1)*loc_s,  b-1)		! loc_s > 0
      = MAX (loc_sect_l + (n-1)*loc_s,  0)		! loc_s < 0
/*}}}*/

Need:  l, n, s, al, as, b, p
----------------------------------------------------------------*/
#endif
/*}}}*/

/*{{{  #define  ABS(m),  MAX(m,n),  MIN(m,n),  CEILING_DIV(n,d)*/
#define  ABS(m)			((m >= 0) ? m : -m)
#define  MAX(m,n)		((m > n) ? m : n)
#define  MIN(m,n)		((m < n) ? m : n)
#define  CEILING_DIV(n,d)	((n + d - 1) / d)
/*}}}*/
{
	int   abs_as,  al,  as,  b,  cycle,  dim,  eta,  l,  loc_l,  loc_n,
	      loc_s,  loc_sect_l,  loc_u,  n,  rank,  s,  this_p;
	
	dim    = *dim_p;
	this_p = *this_p_p;
	cycle  = *cycle_p;
	
	/*{{{  find  l, n, s, al, as, b*/
	rank = ((mapvec [0] & MA_ARRAY_RANK_MASK) >> MA_ARRAY_RANK_SHIFT);
	
	l  = mapvec [ l_subscr (rank, dim) ];
	n  = mapvec [ n_subscr (rank, dim) ];
	s  = mapvec [ s_subscr (rank, dim) ];
	al = mapvec [al_subscr (rank, dim) ];
	as = mapvec [as_subscr (rank, dim) ];
	b  = mapvec [ b_subscr (rank, dim) ];
	/*}}}*/
	
	loc_s = (s * as);
	if ( ! loc_s)		/* an error! */
		loc_s = 1;	/* a simple fix! */

	loc_sect_l = al + (l * as) - (b * this_p);
	
	/*{{{  eta*/
	/*------------------------------------------------------*/
	/* eta = (loc_s > 0) ?					*/
	/*         MAX ((loc_s - loc_sect_l - 1) / loc_s, 0) :	*/
	/*         MAX ((loc_s - loc_sect_l + b) / loc_s, 0);	*/
	/*------------------------------------------------------*/
	eta = (loc_s > 0) ?
	         ((loc_s - loc_sect_l - 1) / loc_s) :
	         ((loc_s - loc_sect_l + b) / loc_s);
	
	eta = MAX (eta, 0);
	/*}}}*/
	
	loc_l = loc_sect_l + (eta * loc_s);
	
	/*{{{  loc_n*/
	/*------------------------------------------------------*/
	/* loc_n = MAX ((loc_u - loc_l + loc_s) / loc_s, 0)	*/
	/* where:						*/
	/* loc_u = (loc_s > 0) ?				*/
	/*            MIN (loc_sect_l + (n-1)*loc_s,  b-1) :	*/
	/*            MAX (loc_sect_l + (n-1)*loc_s,  0);	*/
	/*------------------------------------------------------*/
	loc_u = loc_sect_l + (n-1)*loc_s;
	loc_u = (loc_s > 0) ?
	            MIN (loc_u,  b-1) :
	            MAX (loc_u,  0);
	
	loc_n = (loc_u - loc_l + loc_s) / loc_s;
	loc_n = MAX (loc_n, 0);
	/*}}}*/

	/*{{{  loc_l = (loc_l / |as|) + (cycle * CEILING (b / |as|))*/
	/*-------------------------------------------------------*/
	/* loc_l = (loc_l / |as|) + (cycle * CEILING (b / |as|)) */
	/*-------------------------------------------------------*/
	abs_as = ABS (as);
	loc_l  = (loc_l / abs_as) + (cycle * CEILING_DIV (b, abs_as));
	/*}}}*/
	
	/*-------------------------*/
	/* return results via args */
	/*-------------------------*/
	*eta_p   = eta;
	*loc_l_p = loc_l;
	*loc_n_p = loc_n;
}
/*{{{  #undef ABS, MAX, MIN, CEILING_DIV*/
#undef  ABS
#undef  MAX
#undef  MIN
#undef  CEILING_DIV
/*}}}*/
/*}}}*/
